@@ -1,13 +1,13 @@
 #-----------------------
 # @author: Tony Ribeiro
 # @created: 2019/11/05
-# @updated: 2019/11/05
+# @updated: 2021/06/15
 #
 # @desc: simple Synchronizer implementation, for Learning from states transitions from ANY semantic.
 #   - INPUT: a set of pairs of discrete multi-valued states
 #   - OUTPUT: a logic program with constraints that realize only the input transitions
 #   - THEORY:
-#       - NEW (MLJ 2020?)
+#       - NEW (MLJ 2020)
 #   - COMPLEXITY:
 #       - Variables: exponential
 #       - Values: exponential
@@ -16,10 +16,10 @@
 
 from ..utils import eprint
 from ..objects.rule import Rule
-from ..objects.logicProgram import LogicProgram
-from ..algorithms.gula import GULA
-from ..algorithms.pride import PRIDE
-from ..algorithms.algorithm import Algorithm
+
+from . import Algorithm
+from . import GULA
+from . import PRIDE
 
 import csv
 import numpy as np
@@ -39,12 +39,12 @@ class Synchronizer (Algorithm):
     """
 
     # Enable heuristic: only partial state will be generated for impossible state generation for constraints learning
-    HEURISTIC_PARTIAL_IMPOSSIBLE_STATE = False
+    HEURISTIC_PARTIAL_IMPOSSIBLE_STATE = True
 
     @staticmethod
-    def fit(data, features, targets, complete=True): #variables, values, transitions, conclusion_values=None, complete=True):
+    def fit(dataset, complete=True, verbose=0): #variables, values, transitions, conclusion_values=None, complete=True):
         """
-        Preprocess transitions and learn rules for all variables/values.
+        Preprocess state transitions and learn rules for all variables/values.
 
         Args:
             variables: list of string
@@ -55,8 +55,8 @@ class Synchronizer (Algorithm):
                 state transitions of a dynamic system
 
         Returns:
-            LogicProgram
-                - each rules are minimals
+            CDMVLP
+                - each rules/constraints are minimals
                 - the output set explains/reproduces the input transitions
         """
         #eprint("Start LUST learning...")
@@ -71,9 +71,9 @@ class Synchronizer (Algorithm):
         # 1) Use GULA to learn local possibilities
         #------------------------------------------
         if complete:
-            model = GULA.fit(data, features, targets)#variables, values, transitions, conclusion_values)
+            rules = GULA.fit(dataset)
         else:
-            model = PRIDE.fit(data, features, targets)#variables, values, transitions)
+            rules = PRIDE.fit(dataset)
 
         # 2) Learn constraints
         #------------------------------------------
@@ -88,16 +88,18 @@ class Synchronizer (Algorithm):
         #eprint("positives:\n", positives)
         #eprint("negatives:\n", negatives)
 
-        negatives = np.array([tuple(s1)+tuple(s2) for s1,s2 in data])
+        encoded_data = Algorithm.encode_transitions_set(dataset.data, dataset.features, dataset.targets)
+
+        negatives = np.array([tuple(s1)+tuple(s2) for s1,s2 in encoded_data])
         if len(negatives) > 0:
-            negatives = negatives[np.lexsort(tuple([negatives[:,col] for col in reversed(range(0,len(features)))]))]
+            negatives = negatives[np.lexsort(tuple([negatives[:,col] for col in reversed(range(0,len(dataset.features)))]))]
 
         if complete:
-            constraints = GULA.fit_var_val(features+targets, -1, -1, negatives)
+            constraints = GULA.fit_var_val(dataset.features+dataset.targets, -1, -1, negatives)
         else:
             # Extract occurences of each transition
             next_states = dict()
-            for (i,j) in data:
+            for (i,j) in encoded_data:
                 s_i = tuple(i)
                 s_j = tuple(j)
                 # new init state
@@ -112,7 +114,7 @@ class Synchronizer (Algorithm):
             impossible = set()
             for i in next_states:
                 # Extract all possible value of each variable in next state
-                domains = [set() for var in features]
+                domains = [set() for var in dataset.targets]
 
                 for s in next_states[i][1]:
                     for var in range(0,len(s)):
@@ -139,7 +141,7 @@ class Synchronizer (Algorithm):
             # convert impossible transition for PRIDE input
             positives = [list(i)+list(j) for i,j in impossible]
 
-            constraints = PRIDE.fit_var_val(-1, -1, len(features)+len(targets), positives, negatives)
+            constraints = PRIDE.fit_var_val(-1, -1, len(dataset.features)+len(dataset.targets), positives, negatives)
 
         # DBG
         #eprint("Learned constraints:\n", [r.logic_form(variables, values, None, len(variables)) for r in constraints])
@@ -150,25 +152,25 @@ class Synchronizer (Algorithm):
         if not complete:
             necessary_constraints = constraints
         else:
-            #eprint()
+            # Heuristic: clean constraint with not even a rule for each target condition
             for constraint in constraints:
                 #eprint(features)
                 #eprint(targets)
                 #eprint(constraint, " => ", constraint.logic_form(features+targets,targets))
                 applicable = True
-                for (var,val) in constraint.get_body():
+                for (var,val) in constraint.body:
                     #eprint(var)
                     # Each condition on targets must be achievable by a rule head
-                    if var >= len(features):
-                        head_var = var-len(features)
+                    if var >= len(dataset.features):
+                        head_var = var-len(dataset.features)
                         #eprint(var," ",val)
                         matching_rule = False
                         # The conditions of the rule must be in the constraint
-                        for rule in model.get_rules():
+                        for rule in rules:
                             #eprint(rule)
-                            if rule.get_head_variable() == head_var and rule.get_head_value() == val:
+                            if rule.head_variable == head_var and rule.head_value == val:
                                 matching_conditions = True
-                                for (cond_var,cond_val) in rule.get_body():
+                                for (cond_var,cond_val) in rule.body:
                                     if constraint.has_condition(cond_var) and constraint.get_condition(cond_var) != cond_val:
                                         matching_conditions = False
                                         #eprint("conflict on: ",cond_var,"=",cond_val)
@@ -184,20 +186,80 @@ class Synchronizer (Algorithm):
                     #eprint("OK")
                     necessary_constraints.append(constraint)
 
+        constraints = necessary_constraints
+
+        # Clean remaining constraints
+        # TODO
+        necessary_constraints = []
+        for constraint in constraints:
+            # Get applicables rules
+            compatible_rules = []
+            for (var,val) in constraint.body:
+                #eprint(var)
+                # Each condition on targets must be achievable by a rule head
+                if var >= len(dataset.features):
+                    compatible_rules.append([])
+                    head_var = var-len(dataset.features)
+                    #eprint(var," ",val)
+                    # The conditions of the rule must be in the constraint
+                    for rule in rules:
+                        #eprint(rule)
+                        if rule.head_variable == head_var and rule.head_value == val:
+                            matching_conditions = True
+                            for (cond_var,cond_val) in rule.body:
+                                if constraint.has_condition(cond_var) and constraint.get_condition(cond_var) != cond_val:
+                                    matching_conditions = False
+                                    #eprint("conflict on: ",cond_var,"=",cond_val)
+                                    break
+                            if matching_conditions:
+                                compatible_rules[-1].append(rule)
+
+            # DBG
+            #eprint()
+            #eprint(constraint.logic_form(dataset.features,dataset.targets))
+            #eprint(compatible_rules)
+
+            nb_combinations = np.prod([len(l) for l in compatible_rules])
+            done = 0
+
+            applicable = False
+            for combination in itertools.product(*compatible_rules):
+                done += 1
+                #eprint(done,"/",nb_combinations)
+
+                condition_variables = set()
+                conditions = set()
+                valid_combo = True
+                for r in combination:
+                    for var,val in r.body:
+                        if var not in condition_variables:
+                            condition_variables.add(var)
+                            conditions.add((var,val))
+                        elif (var,val) not in conditions:
+                            valid_combo = False
+                            break
+                    if not valid_combo:
+                        break
+
+                if valid_combo:
+                    #eprint("valid combo: ", combination)
+                    applicable = True
+                    break
+
+            if applicable:
+                necessary_constraints.append(constraint)
 
 
-        output = LogicProgram(features, targets, model.get_rules(), necessary_constraints)#, necessary_constraints)
-
-        return output
+        return rules, necessary_constraints
 
     @staticmethod
     def partial_combinations(states, domains):
         output = []
-        Synchronizer.__partial_combinations(states, domains, [], output)
+        Synchronizer._partial_combinations(states, domains, [], output)
         return output
 
     @staticmethod
-    def __partial_combinations(states, domains, current, output):
+    def _partial_combinations(states, domains, current, output):
         #eprint("current: ", current)
         # full state constructed
         if len(current) >= len(domains):
@@ -225,7 +287,7 @@ class Synchronizer (Algorithm):
                 completed_current = current+[-1 for i in range(0, len(domains)-len(current))]
                 output.append(completed_current)
             else: # complete current state
-                Synchronizer.__partial_combinations(states, domains, current, output)
+                Synchronizer._partial_combinations(states, domains, current, output)
 
             current.pop()
 
